#include <DataIO.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include "FTPClient.h"


TFTPClient::TFTPClient()
{
	fAborted = false;
	fControlEndpoint = 0;
	fWaitTime = 5000000;
}

TFTPClient::~TFTPClient()
{
	this->Abort();
	this->Disconnect();
}

void TFTPClient::Abort()
{
	fAborted = true;
}

uint32 TFTPClient::Reply() const
{
	return fReply;
}

const char *TFTPClient::StrReply() const
{
	return fStrReply.String();
}

const char *TFTPClient::StrError() const
{
	return fStrError.String();
}

status_t TFTPClient::Connect(const char *address, uint16 port)
{
	int err;
	
	struct sockaddr_in addr;
	memset(&addr, 0, sizeof(struct sockaddr_in));
	addr.sin_len = sizeof(struct sockaddr_in);
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	
	if (inet_aton(address, &addr.sin_addr) == 0) {
		struct hostent *host = gethostbyname(address);
		if (!host) {
			err = h_errno;
			fStrError.SetTo(hstrerror(err));
			return err;
		}
		addr.sin_addr = *(struct in_addr *)host->h_addr;
	}
	
	fprintf(stderr, "Connecting %u.%u.%u.%u : %u\n",
		((uint8 *)&addr.sin_addr.s_addr)[0],
		((uint8 *)&addr.sin_addr.s_addr)[1],
		((uint8 *)&addr.sin_addr.s_addr)[2],
		((uint8 *)&addr.sin_addr.s_addr)[3],
		port
		);
	
	if (fControlEndpoint != 0) Disconnect();
	
	fControlEndpoint = socket(AF_INET, SOCK_STREAM, 0);
	if (fControlEndpoint == -1) {
		err = errno;
		fStrError.SetTo(strerror(err));
		return err;
	}
	
	// non-blocking mode
	if ((err = SetBlockingMode(false)) != B_OK) return err;
	
	// connect
	if (connect(fControlEndpoint, (struct sockaddr *)&addr, sizeof(addr)) != 0) {
		err = errno;
		fStrError.SetTo(strerror(err));
		if ((err != 0) && (err != EISCONN) && (err != EINPROGRESS) && (err != EALREADY)) return err;
	}
	
	// 接続後の最初の文字列受信
	if ((err = Readres(fWaitTime)) != B_OK) return err;
	
	
	return B_OK;
}

void TFTPClient::Disconnect()
{
	if (fControlEndpoint != 0) {
		close(fControlEndpoint);
		fprintf(stderr, "Disconnected\n");
	}
	fControlEndpoint = 0;
}

status_t TFTPClient::SetBlockingMode(bool wouldBlock)
{
	int err, block;
	if (wouldBlock) {
		block = 0;
		if (setsockopt(fControlEndpoint, SOL_SOCKET, SO_NONBLOCK, &block, sizeof(block)) < 0) {
			err = errno;
			fStrError.SetTo(strerror(err));
			return err;
		}
	} else {
		block = 1;
		if (setsockopt(fControlEndpoint, SOL_SOCKET, SO_NONBLOCK, &block, sizeof(block)) < 0) {
			err = errno;
			fStrError.SetTo(strerror(err));
			return err;
		}
	}
	return B_OK;
}

bool TFTPClient::IsReceivable(long sec, long msec)
{
	fd_set fdset;
	int re;
	struct timeval timeout;
	
	FD_ZERO(&fdset);
	FD_SET(fControlEndpoint, &fdset);
	
	timeout.tv_sec = sec;
	timeout.tv_usec = msec;
	
	re = select(fControlEndpoint + 1, &fdset, NULL, NULL, &timeout);
	
	return (re == 1) ? true : false;
}

status_t TFTPClient::Readline(BString *line, bigtime_t waitTime)
{
	int32 slen, llen;
	char s[1024];
	char *eol;
	bigtime_t limit = system_time() + waitTime;
	while (!fAborted) {
		eol = (char *)memchr(fPendingString.String(), '\n', fPendingString.Length());
		if (eol) {
			eol++;
			llen = eol - fPendingString.String();
			line->SetTo(fPendingString.String(), llen);
			fPendingString.Remove(0, llen);
			return B_OK;
		} else {
			if (IsReceivable(0, 100)) {
				slen = read(fControlEndpoint, s, sizeof(s));
				if (slen <= 0) {
					// Disconnected
					fStrError.SetTo("Disconnected by remote host.");
					return EPIPE;
				} else {
					fPendingString.Append(s, slen);
					if (fPendingString.Length() > 65536) {
						fStrError.SetTo("Invalid response.");
						return EIO;
					}
					limit = system_time() + waitTime;
				}
			} else {
				// No data
				if (system_time() > limit) {
					// Timeout!
					fStrError.SetTo("Operation timedout.");
					return B_TIMED_OUT;
				}
			}
		}
	}
	fStrError.SetTo("Canceled.");
	return B_CANCELED;
}

status_t TFTPClient::Readres(bigtime_t waitTime)
{
	status_t e;
	BString line;
	fPendingString.SetTo("");
	fStrReply.SetTo("");
	while (true) {
		if ((e = Readline(&line, fWaitTime)) != B_OK) return e;
		fStrReply.Append(line);
		if (line.Length() < 4) {
			fStrError.SetTo("Invalid Response.");
			return EIO;
		}
		if (line.String()[3] != '-') {
			fReply = atoi(line.String());
			break;
		}
	}
	return B_OK;
}

status_t TFTPClient::SendCmd(const char *command, uint32 commandLen)
{
	status_t e;
	ssize_t s;
	fAborted = false;
	s = write(fControlEndpoint, command, commandLen);
	if (s == -1) {
		e = errno;
		fStrError.SetTo(strerror(e));
		return e;
	} else if (s == 0) {
		e = EPIPE;
		fStrError.SetTo(strerror(e));
		return e;
	}
	return B_OK;
}

status_t TFTPClient::USER(const char *userName)
{
	status_t e;
	BString buf("USER ");
	buf << userName << "\r\n";
	e = SendCmd(buf.String(), buf.Length());
	if (e != B_OK) return e;
	return Readres(fWaitTime);
}

status_t TFTPClient::PASS(const char *password)
{
	status_t e;
	BString buf("PASS ");
	buf << password << "\r\n";
	e = SendCmd(buf.String(), buf.Length());
	if (e != B_OK) return e;
	return Readres(fWaitTime);
}

status_t TFTPClient::PASV(struct sockaddr_in *addr)
{
	status_t e;
	const char buf[] = "PASV\r\n";
	char *cp;
	memset(addr, 0, sizeof(struct sockaddr_in));
	unsigned int v[6];
	
	e = SendCmd(buf, strlen(buf));
	if (e != B_OK) return e;
	if ((e = Readres(fWaitTime)) != B_OK) return e;
	
	if ((cp = strchr(fStrReply.String(), '(')) == NULL) {
		fStrError.SetTo("Invalid Response.");
		return B_ERROR;
	}
	cp++;
	
	sscanf(cp, "%u,%u,%u,%u,%u,%u", &v[0], &v[1], &v[2], &v[3], &v[4], &v[5]);
	((uint8 *)&addr->sin_addr.s_addr)[0] = v[0];
	((uint8 *)&addr->sin_addr.s_addr)[1] = v[1];
	((uint8 *)&addr->sin_addr.s_addr)[2] = v[2];
	((uint8 *)&addr->sin_addr.s_addr)[3] = v[3];
	((uint8 *)&addr->sin_port)[0] = v[4];
	((uint8 *)&addr->sin_port)[1] = v[5];
	addr->sin_len = sizeof(struct sockaddr_in);
	addr->sin_family = AF_INET;
	
//	fprintf(stderr, " %d\n", 256 * v[4] + v[5]);
	
	return B_OK;
}

status_t TFTPClient::NLST(struct sockaddr_in *addr, bool passive)
{
	const char buf[] = "NLST -alL\r\n";
	return SendCmd(buf, strlen(buf));
}

status_t TFTPClient::XPWD(BString *result)
{
	status_t e;
	const char buf[] = "XPWD\r\n";
	e = SendCmd(buf, strlen(buf));
	if (e != B_OK) return e;
	if ((e = Readres(fWaitTime)) != B_OK) return e;
	
	BString r(fStrReply.String());
	char *s = strchr(r.String(), '"');
	if (!s) {
		fStrError.SetTo("Invalid response.");
		return B_ERROR;
	}
	s++;
	char *f = strchr(s, '"');
	if (!f) {
		fStrError.SetTo("Invalid response.");
		return B_ERROR;
	}
	*f = 0;
	result->SetTo(s);
	return B_OK;
}

status_t TFTPClient::CWD(const char *dir)
{
	status_t e;
	BString buf("CWD ");
	buf << dir << "\r\n";
	e = SendCmd(buf.String(), buf.Length());
	if (e != B_OK) return e;
	return Readres(fWaitTime);
}
